
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <linux/i2c-dev.h>
#include "mcp23017.h" 
#include "adadisp.h"


/* MCP23017_IODIRB set to 00000000 = 0x00 for writing */
#define IODIRB_SET_WRITE 0x00

/* MCP23017_IODIRB set to 00011110 = 0x1E for reading  */
#define IODIRB_SET_READ 0x1E

/* Initial state of port B: 
	01011111 - Read data mode, ENA to start out at L, 
	Blue (not used anyways) at 1
*/ 
#define OLATB_INIT 0x5F


#define OLATA_INIT 0xff
/* port value: 111.... = 0xFF  */ 

#define IODIRA_INIT 0x1F
/* MCP23017_IODIRA set to 00011111 = 0x1F  here.*/ 

/* command or data line setting */
#define DO_LCD_STATUS 0
#define DO_LCD_COMMAND 0
#define DO_LCD_DATA  0x80


/* LCD commands */

#define LCD_CLEARDISPLAY            0x01
#define LCD_RETURNHOME              0x02
#define LCD_ENTRYMODESET            0x04
#define LCD_DISPLAYCONTROL          0x08
#define LCD_CURSORSHIFT             0x10
#define LCD_FUNCTIONSET             0x20
#define LCD_SETCGRAMADDR            0x40
#define LCD_SETDDRAMADDR            0x80


/* flags for display entry mode */

#define LCD_ENTRYRIGHT              0x00
#define LCD_ENTRYLEFT               0x02
#define LCD_ENTRYSHIFTINCREMENT     0x01
#define LCD_ENTRYSHIFTDECREMENT     0x00


/* flags for display on/off control */

#define LCD_DISPLAYON               0x04
#define LCD_DISPLAYOFF              0x00
#define LCD_CURSORON                0x02
#define LCD_CURSOROFF               0x00
#define LCD_BLINKON                 0x01
#define LCD_BLINKOFF                0x00


/* flags for display/cursor shift */

#define LCD_DISPLAYMOVE             0x08
#define LCD_CURSORMOVE              0x00
#define LCD_MOVERIGHT               0x04
#define LCD_MOVELEFT                0x00


/* flags for function set */

#define LCD_8BITMODE                0x10
#define LCD_4BITMODE                0x00
#define LCD_2LINE                   0x08
#define LCD_1LINE                   0x00
#define LCD_5x10DOTS                0x04
#define LCD_5x8DOTS                 0x00


char *vischar(unsigned int cc, char *buff, size_t buffmax)
{
	if(cc <  0x20 || cc == 127) { buff[0] = '^'; buff[1] = cc^0x40; return buff; } 
	if(cc >= 0x20 && cc <= 126) { buff[0] = cc; buff[1] = 0; return buff; }

	snprintf(buff, buffmax, "%02x", cc); 
	return(buff);
}


/* Write a 4-bit flipped nibble as command or data. */

static int writenibble(int fd, 
		char *what, char *hl, 
		uint8_t cdx,           /* Bit 7 set for command, clear for data */
		uint8_t flipdnibble)
{
	int s; 
	uint8_t wbx, wb0;	
	int rb;

	cdx &= 0x80;   /* Mask away irrelevant and dangerous bits */
	flipdnibble &= 0x1E; /* Mask away the parts not of interest */

	rb = i2c_smbus_read_byte_data(fd, MCP23017_OLATB);
	if(rb == -1) return(-1);
	wb0 = rb & 0x1; /* Retain the low bit, for the blue LED */

	s = i2c_smbus_write_byte_data(fd, MCP23017_IODIRB, 0x00);
	if(s == -1) return(-1);
	// microdelay(50);

	wbx = cdx | flipdnibble | wb0 ;	/* 100....x - write command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);

	/* printf("%s(%d): Write %s %s 0x%02x %c %c: %c %c %c %c\n", 
		__FILE__, __LINE__, what, hl, wbx, 
		wbx & 0x80 ? '1' : '0', 
		wbx & 0x40 ? '1' : '0',
		wbx & 0x2  ? '1' : '0', 
		wbx & 0x4  ? '1' : '0', 
		wbx & 0x8  ? '1' : '0', 
		wbx & 0x10 ? '1' : '0');  */ 

	if(s == -1) return(-1);
	// microdelay(50);

	wbx |= 0x20;                /* 101....x - write command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	wbx &= 0xDF;               /* 100....x - write command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	return(0);
}


/* Read a command or data, return the two bunches of the raw 4 bits */
static int readrcd(int fd, 
		uint8_t cdx,    /* Bit 7 set for command clear for data */
		uint8_t *rfirst, uint8_t *rsecond) 
{
	int s; 
	uint8_t wbx, wb0;	
	int rb;

	cdx &= 0x80;   /* Mask away irrelevant and dangerous bits */
	cdx |= 0x40;  /* set R/W to H for R */ 

	rb = i2c_smbus_read_byte_data(fd, MCP23017_OLATB);
	if(rb == -1) return(-1);
	wb0 = rb & 0x1; /* Retain the low bit, for the blue LED */

	s = i2c_smbus_write_byte_data(fd, MCP23017_IODIRB, IODIRB_SET_READ);
	if(s == -1) return(-1);

	wbx = cdx | 0x1E | wb0 ;	/* 110....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	wbx |= 0x20;                /* 111....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	s = i2c_smbus_read_byte_data(fd, MCP23017_GPIOB);
	if(s == -1) return(-1);
	*rfirst = s; 

	wbx &= 0xDF;               /* 110....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	wbx = cdx | 0x1e | wb0 ;	/* 110....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	wbx |= 0x20;                /* 111....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	s = i2c_smbus_read_byte_data(fd, MCP23017_GPIOB);
	if(s == -1) return(-1);
	*rsecond = s;

	wbx &= 0xDF;               /* 110....x - read command */ 
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, wbx);
	if(s == -1) return(-1);
	// microdelay(50);

	return(0);
}

/* Convert a regular 8-bit value to the two flipped 4-bit values 
	high nibble is to be the first, low nibble is second.  */

/* B1 <> D7 */
/* B2 <> D6 */
/* B3 <> D5 */
/* B4 <> D4 */

static void flip8(uint8_t orig, uint8_t *first, uint8_t *second)
{
	uint8_t vx;
	vx = 0;
	if(orig & 0x80) vx |= 2;
	if(orig & 0x40) vx |= 4;
	if(orig & 0x20) vx |= 8;
	if(orig & 0x10) vx |= 16;
	*first = vx;
	vx = 0;
	if(orig & 0x08) vx |= 2;
	if(orig & 0x04) vx |= 4;
	if(orig & 0x02) vx |= 8;
	if(orig & 0x01) vx |= 16;
	*second = vx;
}


/* Convert the two flipped values read back to the regular byte */
static void unflip8(uint8_t *normal, uint8_t first, uint8_t second)
{
	uint8_t vx;
	vx = 0;

	if(first & 2) vx |= 0x80;
	if(first & 4) vx |= 0x40;
	if(first & 8) vx |= 0x20;
	if(first & 16) vx |= 0x10;

	if(second & 2) vx |= 0x08;
	if(second & 4) vx |= 0x04;
	if(second & 8) vx |= 0x02;
	if(second & 16) vx |= 0x01;

	*normal = vx;
}


/* Write top half of a command for resets. */
static int writehalf(int fd, char *what, uint8_t cmd)
{
	uint8_t first, second;
	int s;
	flip8(cmd, &first, &second);
	s = writenibble(fd, what, "F", DO_LCD_COMMAND, first); 
	return(s);
}

int checkbflag(int fd)
{
	int bf, s; 
	uint8_t first, second;

	s = readrcd(fd, DO_LCD_STATUS, &first, &second); 
	if(s == -1) return(0); /* Don't think the flag is set */
	bf = first & 2; 

	/* Busy flag is bit 7, which we find as bit 1 here */
	/* printf("%s(%d): Read check bf = 0x%x BF = %d\n", 
		__FILE__, __LINE__, first, bf);  */ 

	return(bf); 
}


/* Write a command. Check that the busy flag 
	is clear before writing each half */
static int writecmd(int fd, char *what, uint8_t cmd)
{
	uint8_t first, second;
	int s;
	char info[128];
	snprintf(info, 128, "%s 0x%02x", what, cmd); 

	flip8(cmd, &first, &second);

	/* Check the busy-flag */
	while( checkbflag(fd) != 0);
	s = writenibble(fd, info, "H", DO_LCD_COMMAND, first); 
	s = writenibble(fd, info, "L", DO_LCD_COMMAND, second); 

	return(s);
}


static int writedata(int fd, char *what, uint8_t item)
{
	uint8_t first, second;
	int s;
	char vcbuf[10];
	char info[128];

	vischar(item, vcbuf, 10); 
	snprintf(info, 128, "%s `%s` 0x%02x", what, vcbuf, item); 

	flip8(item, &first, &second);

	/* Check the busy-flag */
	while( checkbflag(fd) != 0);
	s = writenibble(fd, info, "H", DO_LCD_DATA, first); 
	s = writenibble(fd, info, "L", DO_LCD_DATA, second); 

	return(s);
}


int setinitial(int fd)
{
	int s; 

	/* Set the registers in the MCP23017 */
	s = i2c_smbus_write_byte_data(fd, MCP23017_IOCON, 0x00);
	if(s == -1) return(-1);

	s = i2c_smbus_write_byte_data(fd, MCP23017_IODIRB, 0x1E);
	if(s == -1) return(-1);
	
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATB, 0x5f);
	if(s == -1) return(-1);

	s = i2c_smbus_write_byte_data(fd, MCP23017_IODIRA, 0x1F);
	if(s == -1) return(-1);

	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATA, 0xFF);
	if(s == -1) return(-1);

	/* Make the buttons that give L when pressed show as 1. */
	s = i2c_smbus_write_byte_data(fd, MCP23017_IOPOLA, 0x1F);
	if(s == -1) return(-1);

	s = i2c_smbus_write_byte_data(fd, MCP23017_GPPUA, 0x3F);
	if(s == -1) return(-1);

	s = i2c_smbus_write_byte_data(fd, MCP23017_GPPUB, 0x1E);
	if(s == -1) return(-1);

	/* Configure the display itself. */

	/* command-reset */
	delay(16); 				/* Wait for > 15 ms */
	writehalf(fd, "reset 1", 0x30);
	delay(5); 				/* Wait for > 4.1 ms */
	writehalf(fd, "reset 2", 0x30);
	delay(1); 		/* Wait for > 0.1 ms */

	writehalf(fd, "reset 3", 0x30);
	writehalf(fd, "reset 4", 0x20);
	// writecmd(fd, "f-set", 0x32);  /* Double-write */

	/* Now for the real thing. */
	writecmd(fd, "f-set", LCD_FUNCTIONSET | LCD_4BITMODE | LCD_5x8DOTS | LCD_2LINE); 
	// writecmd(fd, "d-ctl", LCD_DISPLAYCONTROL | LCD_DISPLAYON | LCD_CURSORON | LCD_BLINKON);
	writecmd(fd, "d-ctl", LCD_DISPLAYCONTROL | LCD_DISPLAYON);
	writecmd(fd, "e-mst", LCD_ENTRYMODESET | LCD_ENTRYLEFT  | LCD_ENTRYSHIFTDECREMENT);

	cleardisp(fd);

	return(0);
}


int cleardisp(int fd)
{
	int s; 
	s = writecmd(fd, "clear", LCD_CLEARDISPLAY);
	return(s);
}


int backlight(int fd, int blonoff)
{
	int rb, s;
	uint8_t wb0;

	rb = i2c_smbus_read_byte_data(fd, MCP23017_OLATA);
	if(rb == -1) return(-1);
	if(blonoff) wb0 = rb & ~0x40;
	else        wb0 = rb | 0x40;
	s = i2c_smbus_write_byte_data(fd, MCP23017_OLATA, wb0);
	return(s);
}

int gotoca(int fd, int address)
{
	int s;
	address &= 0x7f;
	s = writecmd(fd, "goto", LCD_SETDDRAMADDR | address);
	return(s);
}

int whereis(int fd)
{
	int whs, s; 
	uint8_t hereis, first, second;

	s = readrcd(fd, DO_LCD_STATUS, &first, &second); 
	if(s == -1) return(0); /* Don't think the flag is set */

	unflip8(&hereis, first, second); 
	whs = hereis & 0x7F; 
	return(whs); 
}


/* Put out character and scroll/move to next visible location. */
int ocharmove(int fd, uint8_t thechar)
{
	int whs, s;

	s = writedata(fd, "text", thechar);
	if(s < 0) return(s);

	// 0    0    1    1    2    2    3    3    4    4    5    5    6    6    7
	// 0....5....0....5....0....5....0....5....0....5....0....5....0....5....0
	// abcdefghijklmnop........................ABCDEFGHIJKLMNOP...............

	whs = whereis(fd);

	// printf("%s(%d): Address at:%d\n", __FILE__, __LINE__, whs);

	if(whs >= 16 && whs <= 63) 
	{
		gotoca(fd, 64);	
		return(whs);
	}

	if(whs >= 80 && whs <= 127) 
	{
		gotoca(fd, 0);
	}

	return(whs);
}


int dtext(int fd, char *text)
{
	char *cp;

	for(cp = text; *cp; cp++)
	{
		ocharmove(fd, *cp);
	}
	return(0);

}

/* Read the button-states. */
int readb(int fd)
{
	int s;
	s = i2c_smbus_read_byte_data(fd, MCP23017_GPIOA);
	return(s);
}


#if(0)


int main(int argc, char **argv)
{
	int fd, ba, s;
	char i2cdev[128];
	int j;

	ba = 0x20;
	strcpy(i2cdev, "/dev/i2c-1"); 

	fd = open(i2cdev, O_RDWR);
	if(fd < 0) return(-1); 

	s = ioctl(fd, I2C_SLAVE, ba);
	if(s != 0) 
	{
		close(fd); 
		return(2); 
	}

	setinitial(fd);
	backlight(fd, 1);

	for(j = 1; j<argc; j++)
	{
		dtext(fd, argv[j]);
	}

	s = readb(fd);
	printf("Button state = 0x%x\n", s);

	printf("Buttons: %s %s %s %s %s %s\n", 
		s & ADADISP_BUTTON_S ? "S" : ".", 
		s & ADADISP_BUTTON_L ? "L" : ".",
		s & ADADISP_BUTTON_U ? "U" : ".", 
		s & ADADISP_BUTTON_D ? "D" : ".", 
		s & ADADISP_BUTTON_R ? "R" : ".", 
		""); 

	printf("Buttons: %s %s %s %s %s %s\n", 
		s & 1  ? "S" : ".", 
		s & 16 ? "L" : ".",
		s & 8  ? "U" : ".", 
		s & 4  ? "D" : ".", 
		s & 2  ? "R" : ".", 
		""); 

	// A0 <- Select sw
	// A1 <- Right sw
	// A2 <- Down sw
	// A3 <- Up sw
	// A4 <- Left sw

	close(fd);
	return(0);
}


#endif




